"""Echo task for ScriptEngine."""

from scriptengine.tasks.core import Task, timed_runner
from scriptengine.helpers import terminal_colors as tc


class Echo(Task):
    """Echo task, writes a (coloured) message to stdout
    """

    _required_arguments = ('msg', )

    def __init__(self, arguments):
        Echo.check_arguments(arguments)
        super().__init__(arguments)

    def __str__(self):
        return f'Echo: {getattr(self, "msg", "")}'

    @timed_runner
    def run(self, context):

        log_msg = self.msg.replace('\n', '')
        self.log_info(log_msg[:15]+'...' if len(log_msg) > 18 else log_msg)

        msg = self.getarg('msg', context)
        print(tc.LIGHTBLUE + str(msg) + tc.RESET)



"""Chdir task for ScriptEngine."""

import os

from scriptengine.tasks.core import Task, timed_runner


class Chdir(Task):
    """Chdir task, changes the current working directory
    """
    _required_arguments = ('path', )

    def __init__(self, arguments):
        Chdir.check_arguments(arguments)
        super().__init__(arguments)

    @timed_runner
    def run(self, context):
        path = self.getarg('path', context)
        self.log_info(f"Change path to {path}")
        os.chdir(path)



"""Getenv (base.getenv) and Setenv (base.setenv) tasks for ScriptEngine"""

import os

from scriptengine.tasks.core import Task, timed_runner


class Getenv(Task):
    """ Getenv task, reads environment variables.
    Getenv.run() takes the list of argument name, value pairs, reads the
    environment variables given by the argument values and stores the values of
    the environment variables as context parameters with the argument names.

    For example:
        base.getenv:
            home: HOME
            foo: BAR
    will store $HOME in context['home'] and $BAR in context['foo'].
    """
    @timed_runner
    def run(self, context):
        vars_ = {
            n: os.environ[self.getarg(n, context)]
            for n in vars(self) if not n.startswith('_')
        }
        self.log_info(
            'Read environment variables into context '
            f'({", ".join(vars_.keys())})'
        )
        self.log_debug(vars_)
        context.update(vars_)


class Setenv(Task):
    """Setenv task, sets environment variables from context.
    Setenv.run() takes the list of argument name, value pairs, and sets the
    environemt variables given by the argument names to the values given by the
    argument values. For example,
        base.setenv:
            foo: one
            bar: 2
    will set $foo to "one" and $bar to "2".
    """
    @timed_runner
    def run(self, context):
        vars_ = {
            n: str(self.getarg(n, context))
            for n in vars(self) if not n.startswith('_')
        }
        self.log_info(f'Set environment variables ({", ".join(vars_.keys())})')
        self.log_debug(vars_)
        os.environ.update(vars_)



import sys
from copy import deepcopy

from deepdiff import DeepDiff, Delta

_python_version = (sys.version_info.major, sys.version_info.minor)


def save_copy(context):
    """For versions other than 3.6, this is just copy.deepcopy.
    For 3.6, we have to remove ["se"]["instance"] before deepcopying and restore
    afterwards, because this constitutes some recursive reference that 3.6 has a
    problem with.
    """
    if _python_version != (3, 6):
        return deepcopy(context)
    singleton = object()
    try:
        se_instance = context["se"]["instance"]
        context["se"]["instance"] = None
    except KeyError:
        se_instance = singleton
    copied_context = deepcopy(context)
    if se_instance is not singleton:
        copied_context["se"]["instance"] = context["se"]["instance"] = se_instance
    return copied_context


def context_delta(first, second):
    """deepdiff.Delta.__add__ uses copy.deepcopy internally if Delta was created
    with mutate=False (the default). We set mutate=True for Python 3.6 in order to
    avoid the use of deepcopy (see above).
    """
    if _python_version != (3, 6):
        return Delta(DeepDiff(first, second))
    return Delta(DeepDiff(first, second), mutate=True)



"""ScriptEngine command line interface (cli).

The se command provides a command line interface to ScriptEngine, allowing for
the creation of scripts from YAML files and runnig the scripts via the
SimpleScriptEngine.
"""

__copyright__ = """
Copyright 2019, 2020 Uwe Fladrich

This file is part of ScriptEngine.

ScriptEngine is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

ScriptEngine is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ScriptEngine. If not, see <https://www.gnu.org/licenses/>.
"""
__license__ = "GPLv3+"


import os
import argparse
import logging

import scriptengine.version
import scriptengine.helpers.terminal_colors
import scriptengine.logging

from scriptengine.tasks.core.loader import load as load_tasks
from scriptengine.yaml.parser import parse_file as parse_yaml_file

from scriptengine.engines import SimpleScriptEngine
from scriptengine.exceptions import ScriptEngineParseError, ScriptEngineParseFileError

__version__ = scriptengine.version.__version__


def parse_cmd_line_args():
    """Parses the command line arguments with argparse"""
    arg_parser = argparse.ArgumentParser(
        description="ScriptEngine command line tool",
        epilog="Available ScriptEngine tasks: "
        + ", ".join(load_tasks().keys()),
    )
    arg_parser.add_argument(
        "-V",
        "--version",
        help="show ScriptEngine version and exit",
        action="version",
        version=__version__,
    )
    arg_parser.add_argument(
        "--loglevel",
        help="The minimum level of log messages that is "
        "displayed. For verbose output, use "
        '"debug". For minimal output, use "error" '
        'or "critical".',
        choices=["debug", "info", "warning", "error", "critical"],
    )
    arg_parser.add_argument(
        "--nocolor", help="do not use colored terminal output", action="store_true"
    )
    arg_parser.add_argument("files", help="YAML file(s) to read", nargs="+")

    return arg_parser.parse_args()


def parse_files(logger, files):
    """Parses files and returns a script (list of tasks/jobs)."""
    script = []
    for fname in files:
        # next_script is a Task or Job or list of either
        next_script = parse_yaml_file(fname)
        if isinstance(next_script, list):
            script.extend(next_script)
        else:
            script.append(next_script)

    return script


def main():
    """ScriptEngine command line tool

    The context is passed to the ScriptEngine instance and, later, to all
    tasks. Some context information is already set here in the se command
    line tool:
      context['se']['cli']['cwd']         - the original working directory
      context['se']['cli']['script_path'] - list of paths for all scripts
                                            given at the command line (used
                                            as search path for includes, for
                                            example)
      context['se']['instance']           - a reference to the ScriptEngine
                                            instance that executes the task
      context['se']['tasks']['timing']    - task timing information
                                            (defaults to no timing)
    """

    # Parse command line arguments
    parsed_args = parse_cmd_line_args()

    # Choose colored output unless --nocolor
    scriptengine.helpers.terminal_colors.set_theme(
        "none" if parsed_args.nocolor else "standard"
    )

    # Configure logging with level given by --loglevel
    # or by default with logging.INFO
    scriptengine.logging.configure(
        logging.getLevelName(parsed_args.loglevel.upper())
        if parsed_args.loglevel
        else logging.INFO
    )
    logger = logging.getLogger("se.cli")
    logger.info("Logging configured and started")

    # Log all loaded tasks if debugging
    if logger.level <= logging.DEBUG:
        for name, task in load_tasks().items():
            logger.debug(
                f"Loaded task {name}: {task.__name__} " f"from {task.__module__}"
            )

    # Create script by parsing the files given on the command line
    try:
        script = parse_files(logger, parsed_args.files)
    except ScriptEngineParseFileError:
        logger.critical("Could not read all script files")
        return os.EX_NOINPUT
    except ScriptEngineParseError:
        logger.critical("Could not parse all script files")
        return os.EX_DATAERR

    # Note that the following removal of duplicates does not guarantee the
    # order of entries before Python 3.7! We'll accept this in favour of the
    # easier implementation.
    script_path = tuple(
        dict.fromkeys((os.path.dirname(file) for file in parsed_args.files))
    )

    context = {
        "se": {
            "cli": {
                "cwd": os.getcwd(),
                "script_path": script_path,
            },
            "tasks": {
                "timing": {
                    "mode": None,
                    "logging": None,
                    "timers": {},
                },
            },
            "instance": SimpleScriptEngine(),
        }
    }

    # Call ScriptEngine instance to run the script
    context["se"]["instance"].run(script, context)

    return os.EX_OK



""" SimpleScriptEngine for ScriptEngine.

SimpleScriptEngine is a simplistic script engine. It runs all jobs sequentially
on the local machine, without consideration of job contexts. The
SimpleScriptEngine relies on the Job class to run the actual tasks/jobs.
"""

import logging
import sys
from pprint import pprint

from deepdiff import Delta
from deepmerge import always_merger

from scriptengine.context import save_copy
from scriptengine.exceptions import ScriptEngineStopException, ScriptEngineTaskError


class SimpleScriptEngine:
    def __init__(self):
        self.logger = logging.getLogger(
            "se.instance." + self.__class__.__name__.lower()
        )

    def _guarded_run(self, runner, context):
        try:
            return runner.run(context)
        except AttributeError as e:
            self.log_error(
                "STOPPING SimpleScriptEngine due to internal error: "
                f"Cannot run type {type(runner).__name__}"
            )
            error = e
        except ScriptEngineStopException:
            self.log_info("STOPPING SimpleScriptEngine instance upon request")
            error = None
        except ScriptEngineTaskError as e:
            self.log_error(
                "STOPPING SimpleScriptEngine due to task error in "
                f"{runner.reg_name} id <{runner.shortid}>"
            )
            error = e
        if error:
            if self.logger.getEffectiveLevel() <= logging.DEBUG:
                self.log_error("Last context before error:")
                pprint(context)
                self.log_error("Traceback from error:")
                raise error
            else:
                self.log_error("For more debugging info, re-run with loglevel DEBUG")
        sys.exit()

    def run(self, script, context):
        mycontext = save_copy(context)
        for todo in script if isinstance(script, list) else [script]:
            todo_result = self._guarded_run(todo, mycontext)
            if isinstance(todo_result, dict):
                always_merger.merge(mycontext, todo_result)
            elif isinstance(todo_result, Delta):
                mycontext += todo_result
            else:
                always_merger.merge(
                    mycontext, {"se": {"tasks": {"last_result": todo_result}}}
                )
        return mycontext

    def _log(self, level, msg):
        self.logger.log(level, msg)

    def log_debug(self, msg):
        self._log(logging.DEBUG, msg)

    def log_info(self, msg):
        self._log(logging.INFO, msg)

    def log_warning(self, msg):
        self._log(logging.WARNING, msg)

    def log_error(self, msg):
        self._log(logging.ERROR, msg)

    def log_critical(self, msg):
        self._log(logging.CRITICAL, msg)



"""Context task for ScriptEngine."""

from scriptengine.tasks.core import Task, timed_runner


class Context(Task):
    """Context task, sets/updates the SE context.
    The context task takes name, value pairs as arguments and updates the SE
    context accordingly. Values can be scalars, lists, dictionaries and nested
    combinations thereof.
    Note that Context.run() returns a context *update* and that the higher-level
    running instance (a Job or a ScriptEngine instance) is responsible for
    merging the update with an existing context dict."""

    @timed_runner
    def run(self, context):
        context_update = {
            n: self.getarg(n, context) for n in vars(self) if not n.startswith("_")
        }
        self.log_info(f"Context update: {context_update}")
        return context_update



import os
import codecs
import setuptools


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]


setuptools.setup(
    name='scriptengine',
    version=get_version('scriptengine/version.py'),
    author='Uwe Fladrich',
    author_email='uwe.fladrich@protonmail.com',
    description='A lightweight and extensible framework for executing '
                'scripts written in YAML',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    url='https://github.com/uwefladrich/scriptengine',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
    install_requires=[
        'python-dateutil',
        'deepmerge',
        'deepdiff',
        'PyYAML',
        'jinja2',
    ],
    entry_points={
        'console_scripts': [
            'se = scriptengine.cli.se:main',
        ],
        'scriptengine.tasks': [
            # Short aliases for backward compatibility
            'chdir = scriptengine.tasks.base.chdir:Chdir',
            'command = scriptengine.tasks.base.command:Command',
            'context = scriptengine.tasks.base.context:Context',
            'copy = scriptengine.tasks.base.file:Copy',
            'echo = scriptengine.tasks.base.echo:Echo',
            'exit = scriptengine.tasks.base.exit:Exit',
            'find = scriptengine.tasks.base.find:Find',
            'include = scriptengine.tasks.base.include:Include',
            'link = scriptengine.tasks.base.file:Link',
            'make_dir = scriptengine.tasks.base.file:MakeDir',
            'move = scriptengine.tasks.base.file:Move',
            'remove = scriptengine.tasks.base.file:Remove',
            'task_timer = scriptengine.tasks.base.task_timer:TaskTimer',
            'template = scriptengine.tasks.base.template:Template',
            'time = scriptengine.tasks.base.time:Time',
            # Real names
            'base.chdir = scriptengine.tasks.base.chdir:Chdir',
            'base.command = scriptengine.tasks.base.command:Command',
            'base.context = scriptengine.tasks.base.context:Context',
            'base.copy = scriptengine.tasks.base.file:Copy',
            'base.echo = scriptengine.tasks.base.echo:Echo',
            'base.exit = scriptengine.tasks.base.exit:Exit',
            'base.find = scriptengine.tasks.base.find:Find',
            'base.getenv = scriptengine.tasks.base.envvars:Getenv',
            'base.include = scriptengine.tasks.base.include:Include',
            'base.link = scriptengine.tasks.base.file:Link',
            'base.make_dir = scriptengine.tasks.base.file:MakeDir',
            'base.move = scriptengine.tasks.base.file:Move',
            'base.remove = scriptengine.tasks.base.file:Remove',
            'base.setenv = scriptengine.tasks.base.envvars:Setenv',
            'base.task_timer = scriptengine.tasks.base.task_timer:TaskTimer',
            'base.template = scriptengine.tasks.base.template:Template',
            'base.time = scriptengine.tasks.base.time:Time',
        ]
    }
)



""" ScriptEngine task handling

Tasks are the basic units or work in ScriptEngine. A task serves a single,
specific limited purpose and should require a limited amount or work. Tasks
must have a well defined set of input parameters and be rather independent of
the execution context (e.g. parallel execution).

This module provides
  - the Task base class
  - a timing wrapper for Task.run() functions
"""

from .task import Task
from .timing import timed_runner



""" ScriptEngine jobs.

Jobs are the second level (above tasks) work unit in ScriptEngine. Jobs are
made of tasks and may include loops, conditionals, and a context. Job lists can
be given to a ScriptEngine, for execution.
"""

import ast
import logging
import uuid

from deepdiff import Delta
from deepmerge import always_merger

from scriptengine.context import context_delta, save_copy
from scriptengine.exceptions import ScriptEngineParseError
from scriptengine.jinja import render as j2render
from scriptengine.tasks.core import Task


def _listy(thing, type_=list):
    if isinstance(thing, str) or isinstance(thing, dict):
        return type_((thing,))
    try:
        iter(thing)
    except TypeError:
        return type_((thing,))
    else:
        return type_(thing)


def _is_todo(thing):
    return isinstance(thing, Task) or isinstance(thing, Job)


def _todo_list(todo):
    todo_list = _listy(todo)
    if not all(_is_todo(t) for t in todo_list):
        raise ValueError("A todo list must be made of Tasks and/or Jobs")
    return todo_list


class Job:
    def __init__(self, todo=None, *, when=None, loop=None, loop_vars=None):
        self._identifier = uuid.uuid4()
        self.todo = todo or []
        self._when = when
        self._loop = loop
        self._loop_vars = loop_vars
        self.log_debug(
            "Create Job"
            f'{" when "+self._when if self._when else ""}'
            f'{" with "+str(self._loop_vars) if self._loop else ""}'
            f'{" in "+str(self._loop) if self._loop else ""}'
        )

    @property
    def id(self):
        return self._identifier

    @property
    def shortid(self):
        return self._identifier.hex[:10]

    @property
    def todo(self):
        return self._todo

    @todo.setter
    def todo(self, todo):
        self._todo = _todo_list(todo)

    def when(self, context):
        return self._when is None or j2render(self._when, context, boolean=True)

    def loop(self, context):
        if self._loop:
            loop_iter = j2render(self._loop, context)
            if isinstance(loop_iter, str):
                try:
                    loop_iter = ast.literal_eval(loop_iter or "None")
                except SyntaxError:
                    self.log_error(
                        f"Syntax error while evaluating loop expression '{loop_iter}'"
                    )
                    raise ScriptEngineParseError
            if isinstance(loop_iter, dict):
                loop_iter = loop_iter.items()
                loop_vars = self._loop_vars or ("key", "value")
            else:
                loop_vars = self._loop_vars or ("item",)
            for items in loop_iter:
                yield dict(
                    zip(
                        _listy(loop_vars, type_=tuple),
                        (j2render(i, context) for i in _listy(items, type_=tuple)),
                    )
                )
        else:
            yield {}

    def append(self, todo):
        todo_list = _todo_list(todo)
        self._todo.extend(todo_list)
        self.log_debug(f'Append: {",".join(t.shortid for t in todo_list)}')

    def run(self, context):
        if self.when(context):
            job_context = save_copy(context)
            for items in self.loop(job_context):
                if set(items) & set(job_context):
                    self.log_warning(
                        "The following loop variables collide with the "
                        f"context: {set(items) & set(job_context)}"
                    )
                for t in self.todo:
                    todo_result = t.run({**job_context, **items})
                    if isinstance(todo_result, dict):
                        always_merger.merge(job_context, todo_result)
                    elif isinstance(todo_result, Delta):
                        job_context += todo_result
                    else:
                        always_merger.merge(
                            job_context, {"se": {"tasks": {"last_result": todo_result}}}
                        )
            return context_delta(context, job_context)

    def _log(self, level, msg):
        logging.getLogger("se.job").log(level, msg, extra={"id": self.shortid})

    def log_debug(self, msg):
        self._log(logging.DEBUG, msg)

    def log_info(self, msg):
        self._log(logging.INFO, msg)

    def log_warning(self, msg):
        self._log(logging.WARNING, msg)

    def log_error(self, msg):
        self._log(logging.ERROR, msg)

    def log_critical(self, msg):
        self._log(logging.CRITICAL, msg)



"""ScriptEngine file tasks
Provides Copy, Move, Link, Remove and MakeDir tasks
"""

import pathlib
import shutil

from scriptengine.tasks.core import Task, timed_runner
from scriptengine.exceptions import ScriptEngineTaskRunError


def _getarg_path(self, name, context={}, **kwargs):
    """Wrapper for Task.getarg() that returns a pathlib.Path"""
    arg = self.getarg(name, context, **kwargs)
    try:
        return pathlib.Path(arg)
    except TypeError:
        self.log_error(
            f'Invalid "{name}" argument, "{arg}" is not a path'
        )
        raise ScriptEngineTaskRunError


class Copy(Task):
    """ScriptEngine Copy task: Copies files or directories. It needs 'src' and
    'dst' parameters when it is created, providing the source and destination
    paths for the copy operation. Directories are copied recursively,
    maintaining symlinks.
    """
    _required_arguments = ('src', 'dst', )

    def __init__(self, arguments):
        Copy.check_arguments(arguments)
        super().__init__(arguments)

    @timed_runner
    def run(self, context):

        src = _getarg_path(self, 'src', context)
        dst = _getarg_path(self, 'dst', context)

        if src.is_file():
            self.log_info(f'Copy file: {src} --> {dst}')
            if dst.exists():
                self.log_warning(
                    f'Destination file "{dst}" exists already; overwriting')
            shutil.copy(src, dst)
        elif src.is_dir():
            self.log_info(f'Copy directory: {src} --> {dst}')
            if dst.exists():
                self.log_error(
                    f'Target directory "{dst}" exists'
                )
                raise ScriptEngineTaskRunError
            shutil.copytree(str(src), str(dst), symlinks=True)
        elif not src.exists():
            ignore = self.getarg('ignore_not_found', context, default=False)
            if ignore:
                self.log_warning(
                    f'"src" does not exist: {src}'
                )
            else:
                self.log_error(
                    f'"src" does not exist: {src}'
                )
                raise ScriptEngineTaskRunError
        else:
            self.log_error(
                f'"src" is not a file or directory: {src}'
            )
            raise ScriptEngineTaskRunError


class Move(Task):
    """ScriptEngine Move task: Moves files or directories. It needs 'src' and
    'dst' parameters when it is created, providing the source and destination
    paths for the move operation.
    """
    _required_arguments = ('src', 'dst', )

    def __init__(self, arguments):
        Move.check_arguments(arguments)
        super().__init__(arguments)

    @timed_runner
    def run(self, context):
        src = _getarg_path(self, 'src', context)
        dst = _getarg_path(self, 'dst', context)
        self.log_info(f'Move file: {src} --> {dst}')
        shutil.move(str(src), str(dst))


class Link(Task):
    """ScriptEngine Link task: Creates symlinks. It needs 'src' and 'dst'
    parameters when it is created, providing the source and destination paths
    for the link.
    """
    _required_arguments = ('src', 'dst', )

    def __init__(self, arguments):
        Link.check_arguments(arguments)
        super().__init__(arguments)

    @timed_runner
    def run(self, context):
        src = _getarg_path(self, 'src', context)
        dst = _getarg_path(self, 'dst', context)
        self.log_info(f'Create link: {src} --> {dst}')
        dst.symlink_to(src)
        if not dst.exists():
            self.log_warning(f'Created dangling symlink: {dst}-->{src}')


class Remove(Task):
    """ScriptEngine Remove task: Removes (deletes) files or directories.
    Directories are removed recursively. It needs the 'path' parameter when it
    is created, providing the path to the file or directory to be removed.
    """
    _required_arguments = ('path', )

    def __init__(self, arguments):
        Remove.check_arguments(arguments)
        super().__init__(arguments)

    @timed_runner
    def run(self, context):
        path = _getarg_path(self, 'path', context)
        if path.exists():
            if path.is_dir():
                self.log_info(f'Removing directory "{path}"')
                shutil.rmtree(str(path))
            else:
                self.log_info(f'Removing "{path}"')
                path.unlink()
        else:
            self.log_info(f'Non-existing path "{path}"; nothing removed')


class MakeDir(Task):
    """ScriptEngine MakeDir task: Creates a directory. It needs the 'path'
    parameter when it is created, providing the path of the new directory
    """
    _required_arguments = ('path', )

    def __init__(self, arguments):
        MakeDir.check_arguments(arguments)
        super().__init__(arguments)

    @timed_runner
    def run(self, context):
        path = _getarg_path(self, 'path', context)
        try:
            path.mkdir(parents=True)
        except FileExistsError:
            if path.is_dir():
                self.log_info(
                    f'Directory "{path}" exists already'
                )
            else:
                self.log_error(
                    f'A file in the way of creating directory "{path}"'
                )
                raise ScriptEngineTaskRunError
        else:
            self.log_info(
                f'Created directory "{path}"'
            )



import pkg_resources
import logging
import functools

from scriptengine.exceptions import ScriptEngineTaskLoaderError


# The load function goest through entry points, searching for ScriptEngine
# tasks. Since the function can be called quite frequently and modules are
# *usually* not loaded while SE is run, the result is cached.
# Note that this means that dynamic module is *not supported* by the loading
# mechanism!
@functools.lru_cache(maxsize=None)
def load():

    entry_point = 'scriptengine.tasks'

    loaded_tasks = dict()
    for ep in pkg_resources.iter_entry_points(entry_point):
        if ep.name not in loaded_tasks:
            loaded_tasks[ep.name] = ep.load()
        else:
            clash = next(clash_ep.module_name
                         for clash_ep
                         in pkg_resources.iter_entry_points(entry_point)
                         if clash_ep.name == ep.name)
            logging.getLogger('se.task.loader').error(
                f'Same task name "{ep.name} defined in modules '
                f'"{ep.module_name}" and "{clash}"'
            )
            raise ScriptEngineTaskLoaderError

    return loaded_tasks


def load_and_register():
    loaded_tasks = load()
    for name, task in loaded_tasks.items():
        task.register_name(name)
    return loaded_tasks



"""Command task for ScriptEngine

   Example:
      - command:
            name: ls
            args: [-l, -a]
"""

import subprocess
import threading
import os
from contextlib import contextmanager

from scriptengine.tasks.core import Task, timed_runner
from scriptengine.exceptions import (
    ScriptEngineTaskRunError,
    ScriptEngineTaskArgumentInvalidError,
)


class LogPipe(threading.Thread):
    """Helper class that can be used in the subprocess.run() call as argument
    for stdout and stderr. It implements a pipe that sends whatever it receives
    to a log function (provided at initialisation). This can be used to send
    stdout/stderr of a subprocess to a logger.
    Follows https://codereview.stackexchange.com/questions/6567"""

    def __init__(self, log_function):
        super().__init__()
        self.log_function = log_function
        self.fd_read, self.fd_write = os.pipe()
        self.pipe_reader = os.fdopen(self.fd_read)
        self.start()

    def fileno(self):
        return self.fd_write

    def run(self):
        for line in iter(self.pipe_reader.readline, ""):
            self.log_function(line.strip("\n"))
        self.pipe_reader.close()

    def close(self):
        os.close(self.fd_write)


class Command(Task):
    """Command task, executes a command in a shell"""

    _required_arguments = ("name",)

    def __init__(self, arguments):
        Command.check_arguments(arguments)
        super().__init__(arguments)

    @timed_runner
    def run(self, context):
        self.log_info(
            f"{self.name} "
            f'args={getattr(self, "args", None)} '
            f'cwd={getattr(self, "cwd", None)}'
        )

        command = self.getarg("name", context)

        args = self.getarg("args", context, default=[])
        args = args if isinstance(args, list) else [args]

        cwd = self.getarg("cwd", context, default=None)
        ignore_error = self.getarg("ignore_error", context, default=False)

        self.log_debug(
            f"{command} "
            f'{" ".join(map(str, args))} '
            f"cwd={cwd} "
            f"ignore_error={ignore_error} "
        )

        @contextmanager
        def log_pipe(mode):
            """Returns something that can be used as stdout/stderr argument for
            subprocess.run(). If mode is True, a LogPipe is returned, which
            sends stdout/stderr through the info logger of the tasks. If mode is
            None/False, then None is returned and stdout/stderr are ignored. If
            mode is a string, stdout/stderr is captured by subprocess and later
            stored in the context."""
            if mode is True:
                pipe = LogPipe(self.log_info)
            elif not mode:
                pipe = None
            elif isinstance(mode, str):
                pipe = subprocess.PIPE
            else:
                self.log_error(f"Invalid task argument: {mode}")
                raise ScriptEngineTaskArgumentInvalidError
            try:
                yield pipe
            finally:
                if isinstance(pipe, LogPipe):
                    pipe.close()

        stdout_mode = self.getarg("stdout", context, default=True)
        self.log_debug(
            "stdout mode: "
            f'{stdout_mode if stdout_mode in (True, False) else "context"}'
        )
        stderr_mode = self.getarg("stderr", context, default=True)
        self.log_debug(
            "stderr mode: "
            f'{stderr_mode if stderr_mode in (True, False) else "context"}'
        )
        with log_pipe(stdout_mode) as stdout, log_pipe(stderr_mode) as stderr:
            try:
                cmd_proc = subprocess.run(
                    map(str, (command, *args)),
                    stdout=stdout,
                    stderr=stderr,
                    cwd=cwd,
                    check=True,
                    errors="replace",
                )
            except subprocess.CalledProcessError as e:
                if ignore_error:
                    self.log_warning(f"Command returned error code {e.returncode}")
                else:
                    self.log_error(f"Command returned error code {e.returncode}")
                    raise ScriptEngineTaskRunError
            else:
                if isinstance(stdout_mode, str):
                    context[stdout_mode] = cmd_proc.stdout.split("\n")
                if isinstance(stderr_mode, str):
                    context[stderr_mode] = cmd_proc.stderr.split("\n")



"""ScriptEngine helpers: Jinja2 rendering"""

import jinja2
import os
import datetime
import distutils.util

from scriptengine.exceptions import ScriptEngineParseJinjaError


def string_to_datetime(string, format="%Y-%m-%d %H:%M:%S"):
    """Jinja2 filter to convert a string to datetime.datetime"""
    return datetime.datetime.strptime(string, format)


def string_to_date(string, format="%Y-%m-%d %H:%M:%S"):
    """Jinja2 filter to convert a string to datetime.date"""
    return datetime.datetime.strptime(string, format).date()


def basename(path):
    """Jinja2 filter that returns a path's base name"""
    return os.path.basename(path)


def dirname(path):
    """Jinja2 filter that returns a path's dir name"""
    return os.path.dirname(path)


def filters():
    """Return all defined Jinja2 filters by their name and corresponding function
    """
    return {
        'datetime': string_to_datetime,
        'date': string_to_date,
        'basename': basename,
        'dirname': dirname,
    }


# Jinja2 Environment to be used for rendering of parameters in the YAML files
_param_env = jinja2.Environment(loader=jinja2.BaseLoader)
for name, function in filters().items():
    _param_env.filters[name] = function


def render(arg, context, recursive=True, boolean=False):
    """ Renders a string with Jinja2.

    The argument is rendered via jinja2.Template().render() if it is a string,
    or returned unchanged otherwise. Rendering is done either once or
    recursively until no further variables can be substituted. The result is
    returned either as string, or converted into a bool (True/False).

    Args:
        arg: The string to be rendered. If it is not a string, arg is returned
            without changes.
        recursive (bool): If true (default), the string is rendered
            recursively, i.e. until it's value doesn't change anymore. If
            false, the string is rendered once.
        boolean (bool): If false (default) the rendered string is returned
            (i.e. the return value is a string). If true, the rendered
            string is evaluated in a boolean context in Jinja2.  The return
            value is true or false in this case.
        context (dict): A dictionary that provides the context for
            jinja2.Template().render()

    Returns:
        str: The rendered string (of boolean==False), OR
        bool: The rendered string, evaluated as bool (if boolean=True)
    """

    def render_with_context(string_arg):
        try:
            # Render string in parameter environment using context
            return _param_env.from_string(string_arg).render(context)
        except jinja2.TemplateSyntaxError:
            raise ScriptEngineParseJinjaError(
                        'Syntax error while rendering template string '
                        f'"{string_arg}"'
                        f'{" in boolean context" if boolean else ""}')

    if isinstance(arg, str):
        rendered_string = render_with_context(arg)
        if recursive:
            next_rendered_string = render_with_context(rendered_string)
            while rendered_string != next_rendered_string:
                rendered_string = next_rendered_string
                next_rendered_string = render_with_context(rendered_string)
        if boolean:
            expr = f'{{% if {rendered_string} %}}1{{% else %}}0{{% endif %}}'
            return bool(
                distutils.util.strtobool(
                    render_with_context(expr)
                )
            )
        return rendered_string
    else:
        return arg
